/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             create.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ntifs.h"
#include "ext2fs.h"

/* GLOBALS *****************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2LookupFileName)
#pragma alloc_text(PAGE, Ext2SearchFcbList)
#pragma alloc_text(PAGE, Ext2ScanDir)
#pragma alloc_text(PAGE, Ext2CreateFile)
#pragma alloc_text(PAGE, Ext2CreateVolume)
#pragma alloc_text(PAGE, Ext2Create)
#pragma alloc_text(PAGE, Ext2CreateInode)
#pragma alloc_text(PAGE, Ext2SupersedeOrOverWriteFile)
#endif

NTSTATUS
Ext2LookupFileName (IN PEXT2_VCB    Vcb,
            IN PUNICODE_STRING      FullFileName,
            IN PEXT2_MCB            ParentMcb,
            OUT PEXT2_MCB *         Ext2Mcb,
            IN OUT PEXT2_INODE      Inode)
{
    NTSTATUS        Status;
    UNICODE_STRING  FileName;
    PEXT2_MCB       Mcb;

    EXT2_DIR_ENTRY2 ext2_dir;
    int             i = 0;
    BOOLEAN         bRun = TRUE;
    BOOLEAN         bParent = FALSE;
    EXT2_INODE      in;
    ULONG           off = 0;

    Status = STATUS_OBJECT_NAME_NOT_FOUND;

    *Ext2Mcb = NULL;

    if (ParentMcb) {
        bParent = TRUE;
    } else if (FullFileName->Buffer[0] == L'\\') {
        ParentMcb = Vcb->McbTree;
    } else {
        return STATUS_OBJECT_PATH_NOT_FOUND;
    }

    RtlZeroMemory(&ext2_dir, sizeof(EXT2_DIR_ENTRY2));

    if (FullFileName->Length == 0) {
        return Status;
    }

    if (FullFileName->Length == 2 && FullFileName->Buffer[0] == L'\\') {
        if (!Ext2LoadInode(Vcb, ParentMcb->Inode, Inode))  {
            return Status;      
        }

        *Ext2Mcb = Vcb->McbTree;

        return STATUS_SUCCESS;
    }

    while (bRun && i < FullFileName->Length/2) {
        int Length;
        ULONG FileAttr = FILE_ATTRIBUTE_NORMAL;

        if (bParent) {
            bParent = FALSE;
        } else  {
            while(i < FullFileName->Length/2 && FullFileName->Buffer[i] == L'\\') i++;
        }

        Length = i;

        while(i < FullFileName->Length/2 && (FullFileName->Buffer[i] != L'\\')) i++;

        if (i - Length >0) {
            FileName = *FullFileName;
            FileName.Buffer += Length;
            FileName.Length = (USHORT)((i - Length) * 2);

            Mcb = Ext2SearchMcb(Vcb, ParentMcb, &FileName);

            if (Mcb) {
                ParentMcb = Mcb;

                Status = STATUS_SUCCESS;

                if (!IsFlagOn(Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY)) {
                    if (i < FullFileName->Length/2) {
                        Status = STATUS_OBJECT_PATH_NOT_FOUND;
                    }

                    break;
                }
            } else {
                if (!Ext2LoadInode(Vcb, ParentMcb->Inode, &in)) {
                    Status = STATUS_OBJECT_NAME_NOT_FOUND;
                    break;
                }

                if (!S_ISDIR(in.i_mode)) {
                    if (i < FullFileName->Length/2) {
                        Status =  STATUS_OBJECT_NAME_NOT_FOUND;
                        break;
                    }
                }

                Status = Ext2ScanDir (
                            Vcb,
                            ParentMcb,
                            &FileName,
                            &off,
                            &in,
                            &ext2_dir);

                if (!NT_SUCCESS(Status)) {
                    bRun = FALSE;
/*
                    if (i >= FullFileName->Length/2)
                    {
                        *Ext2Mcb = ParentMcb;
                    }
*/
                } else {
                    if (IsFlagOn( SUPER_BLOCK->s_feature_incompat, 
                                  EXT2_FEATURE_INCOMPAT_FILETYPE)) {
                        if (ext2_dir.file_type == EXT2_FT_DIR)
                            SetFlag(FileAttr, FILE_ATTRIBUTE_DIRECTORY);
                    } else {
                        if (!Ext2LoadInode(Vcb, ext2_dir.inode, &in)) {
                            Status = STATUS_OBJECT_NAME_NOT_FOUND;
                            break;
                        }
                        if (S_ISDIR(in.i_mode)) {
                            SetFlag(FileAttr, FILE_ATTRIBUTE_DIRECTORY);
                        }
                    }

                    SetFlag(ParentMcb->Flags, MCB_IN_USE);
                    Mcb = Ext2AllocateMcb(Vcb, &FileName, FileAttr);
                    ClearFlag(ParentMcb->Flags, MCB_IN_USE);

                    if (!Mcb) {
                        Status = STATUS_OBJECT_NAME_NOT_FOUND;
                        break;
                    }

                    Mcb->Inode = ext2_dir.inode;
                    Mcb->DeOffset = off;
                    Ext2AddMcbNode(Vcb, ParentMcb, Mcb);
                    ParentMcb = Mcb;
                }
            }
        } else {
            break;
        }
    }

    if (NT_SUCCESS(Status)) {
        *Ext2Mcb = Mcb;
        if (Inode) {
            if (!Ext2LoadInode(Vcb, Mcb->Inode, Inode)) {
                Ext2Print((DBG_ERROR, "Ext2LoopupFileName: error loading Inode %xh\n",
                          Mcb->Inode));
                Status = STATUS_INSUFFICIENT_RESOURCES;
            }
        }
    }

    return Status;
}


NTSTATUS
Ext2ScanDir (IN PEXT2_VCB       Vcb,
         IN PEXT2_MCB           ParentMcb,
         IN PUNICODE_STRING     FileName,
         IN OUT PULONG          Index,
         IN PEXT2_INODE         Inode,
         IN OUT PEXT2_DIR_ENTRY2 ext2_dir)
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    USHORT                  InodeFileNameLength;
    UNICODE_STRING          InodeFileName;

    PEXT2_DIR_ENTRY2        pDir = NULL;
    ULONG                   dwBytes = 0;
    BOOLEAN                 bFound = FALSE;
    LONGLONG                Offset = 0;
    ULONG                   inode = ParentMcb->Inode;

    ULONG                   dwRet;

    __try {

        pDir = (PEXT2_DIR_ENTRY2) ExAllocatePool(PagedPool,
                                    sizeof(EXT2_DIR_ENTRY2));
        if (!pDir) {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        InodeFileName.Buffer = ExAllocatePool(
                    PagedPool,
                    (EXT2_NAME_LEN + 1) * 2 );

        if (!InodeFileName.Buffer) {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        dwBytes = 0;

        while (!bFound && dwBytes < Inode->i_size) {

            RtlZeroMemory(pDir, sizeof(EXT2_DIR_ENTRY2));

            //
            // reading dir entries from Dcb
            //

            Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        ParentMcb->Inode,
                        Inode,
                        dwBytes,
                        (PVOID)pDir,
                        sizeof(EXT2_DIR_ENTRY2),
                        &dwRet);


            if (!NT_SUCCESS(Status)) {
                Ext2Print((DBG_ERROR, "Ext2ScanDir: Reading Directory Content error.\n"));
                __leave;
            }

            if (pDir->inode && (pDir->inode < INODES_COUNT)) {

                OEM_STRING OemName;
                OemName.Buffer = pDir->name;
                OemName.Length = (pDir->name_len & 0xff);
                OemName.MaximumLength = OemName.Length;

                InodeFileNameLength = (USHORT)
                        Ext2OEMToUnicodeSize(&OemName);
                
                InodeFileName.Length = 0;
                InodeFileName.MaximumLength = (EXT2_NAME_LEN + 1) * 2;

                RtlZeroMemory( InodeFileName.Buffer,
                               InodeFileNameLength + 2);
                
                Status = Ext2OEMToUnicode(
                                &InodeFileName,
                                &OemName    );

                if (!NT_SUCCESS(Status)) {
                    __leave;
                }

                if (!RtlCompareUnicodeString(
                        FileName,
                        &InodeFileName,
                        TRUE ))  {

                    bFound = TRUE;
                    *Index = dwBytes;
                    RtlCopyMemory(ext2_dir, pDir, pDir->rec_len > sizeof(EXT2_DIR_ENTRY2)
                        ? sizeof(EXT2_DIR_ENTRY2) : pDir->rec_len);
                    Status = STATUS_SUCCESS;

                    Ext2Print(( DBG_INFO, "Ext2ScanDir: Found: Name=%S Inode=%xh\n",
                                InodeFileName.Buffer, pDir->inode ));
                }
                
                dwBytes +=pDir->rec_len;
                Offset = (LONGLONG)dwBytes;

            } else {

                if (pDir->inode == 0) {

                    if (pDir->rec_len == 0) {
                        DbgBreak();
                        break;
                    } else {
                        dwBytes +=pDir->rec_len;
                        Offset = (LONGLONG)dwBytes;
                    }
                } else {
                    break;
                }
            }
        }

        if (!bFound) {
            Status = STATUS_NO_SUCH_FILE;
        }

    } __finally {

        if (InodeFileName.Buffer != NULL) {
            ExFreePool(InodeFileName.Buffer);
        }

        if (pDir)
            ExFreePool(pDir);
    }
    
    return Status;
}


NTSTATUS
Ext2CreateFile(PEXT2_IRP_CONTEXT IrpContext, PEXT2_VCB Vcb)
{
    NTSTATUS            Status = STATUS_UNSUCCESSFUL;
    PIO_STACK_LOCATION  IrpSp;
    PEXT2_FCB           Fcb = NULL;
    PEXT2_MCB           Ext2Mcb = NULL;

    PEXT2_FCB           ParentFcb = NULL;
    PEXT2_MCB           ParentMcb = NULL;

    BOOLEAN             bParentFcbCreated = FALSE;

    PEXT2_CCB           Ccb = NULL;
    PEXT2_INODE         Inode;
    BOOLEAN             VcbResourceAcquired = FALSE;
    BOOLEAN             bDir = FALSE;
    BOOLEAN             bFcbAllocated = FALSE;
    BOOLEAN             bCreated = FALSE;
    UNICODE_STRING      FileName;
    PIRP                Irp;

    ULONG               Options;
    ULONG               CreateDisposition;

    BOOLEAN             OpenDirectory;
    BOOLEAN             OpenTargetDirectory;
    BOOLEAN             CreateDirectory;
    BOOLEAN             SequentialOnly;
    BOOLEAN             NoIntermediateBuffering;
    BOOLEAN             IsPagingFile;
    BOOLEAN             DirectoryFile;
    BOOLEAN             NonDirectoryFile;
    BOOLEAN             NoEaKnowledge;
    BOOLEAN             DeleteOnClose;
    BOOLEAN             TemporaryFile;
    BOOLEAN             CaseSensitive;

    ACCESS_MASK         DesiredAccess;
    ULONG               ShareAccess;


    Irp = IrpContext->Irp;
    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    Options  = IrpSp->Parameters.Create.Options;
    
    DirectoryFile = IsFlagOn(Options, FILE_DIRECTORY_FILE);
    OpenTargetDirectory = IsFlagOn(IrpSp->Flags, SL_OPEN_TARGET_DIRECTORY);

    NonDirectoryFile = IsFlagOn(Options, FILE_NON_DIRECTORY_FILE);
    SequentialOnly = IsFlagOn(Options, FILE_SEQUENTIAL_ONLY);
    NoIntermediateBuffering = IsFlagOn( Options, FILE_NO_INTERMEDIATE_BUFFERING );
    NoEaKnowledge = IsFlagOn(Options, FILE_NO_EA_KNOWLEDGE);
    DeleteOnClose = IsFlagOn(Options, FILE_DELETE_ON_CLOSE);

    CaseSensitive = IsFlagOn(IrpSp->Flags, SL_CASE_SENSITIVE);

    TemporaryFile = IsFlagOn(IrpSp->Parameters.Create.FileAttributes,
                                   FILE_ATTRIBUTE_TEMPORARY );

    CreateDisposition = (Options >> 24) & 0x000000ff;

    IsPagingFile = IsFlagOn(IrpSp->Flags, SL_OPEN_PAGING_FILE);

    CreateDirectory = (BOOLEAN)(DirectoryFile &&
                                ((CreateDisposition == FILE_CREATE) ||
                                 (CreateDisposition == FILE_OPEN_IF)));

    OpenDirectory   = (BOOLEAN)(DirectoryFile &&
                                ((CreateDisposition == FILE_OPEN) ||
                                 (CreateDisposition == FILE_OPEN_IF)));

    DesiredAccess = IrpSp->Parameters.Create.SecurityContext->DesiredAccess;
    ShareAccess   = IrpSp->Parameters.Create.ShareAccess;

    FileName.Buffer = NULL;

    __try {

        ExAcquireResourceExclusiveLite(
            &Vcb->MainResource, TRUE );
        
        VcbResourceAcquired = TRUE;

        if (Irp->Overlay.AllocationSize.HighPart) {
            Status = STATUS_INVALID_PARAMETER;
            __leave;
        }
        
        if (!(Inode = ExAllocatePoolWithTag (
              PagedPool, sizeof(EXT2_INODE), 'EInd' ) )) {
            __leave;
        }

        RtlZeroMemory(Inode, sizeof(EXT2_INODE));

        FileName.MaximumLength = IrpSp->FileObject->FileName.MaximumLength;
        FileName.Length = IrpSp->FileObject->FileName.Length;

        FileName.Buffer = ExAllocatePool(PagedPool, FileName.MaximumLength);
        if (!FileName.Buffer) {   
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        RtlZeroMemory(FileName.Buffer, FileName.MaximumLength);
        RtlCopyMemory(FileName.Buffer, IrpSp->FileObject->FileName.Buffer, FileName.Length);

        if (IrpSp->FileObject->RelatedFileObject) {
            ParentFcb = (PEXT2_FCB)(IrpSp->FileObject->RelatedFileObject->FsContext);
        }

        if ((FileName.Length > sizeof(WCHAR)) &&
            (FileName.Buffer[1] == L'\\') &&
            (FileName.Buffer[0] == L'\\')) {
            
            FileName.Length -= sizeof(WCHAR);
            
            RtlMoveMemory( &FileName.Buffer[0],
                &FileName.Buffer[1],
                FileName.Length );
            
            //
            //  Bad Name if there are still beginning backslashes.
            //
            
            if ((FileName.Length > sizeof(WCHAR)) &&
                (FileName.Buffer[1] == L'\\') &&
                (FileName.Buffer[0] == L'\\')) {
                                
                Status = STATUS_OBJECT_NAME_INVALID;

                __leave;
            }
        }

        if (IsFlagOn(Options, FILE_OPEN_BY_FILE_ID)) {
            Status = STATUS_NOT_IMPLEMENTED;
            __leave;
        }

        Ext2Print((DBG_INFO, "Ext2CreateFile: %S (NameLen=%xh) Paging=%xh Option: %xh.\n",
            FileName.Buffer, FileName.Length, IsPagingFile, IrpSp->Parameters.Create.Options));

        if (ParentFcb) {
            ParentMcb = ParentFcb->Ext2Mcb;
        }

        Status = Ext2LookupFileName(
                        Vcb,
                        &FileName,
                        ParentMcb,
                        &Ext2Mcb,
                        Inode );

        if (!NT_SUCCESS(Status)) {
            UNICODE_STRING  PathName;
            UNICODE_STRING  RealName;
            UNICODE_STRING  RemainName;

            LONG            i = 0;


            PathName = FileName;

            Ext2Print((DBG_INFO, "Ext2CreateFile: File %S will be created.\n", PathName.Buffer));

            Ext2Mcb = NULL;

            if (PathName.Buffer[PathName.Length/2 - 1] == L'\\') {
                if (DirectoryFile) {
                    PathName.Length -=2;
                    PathName.Buffer[PathName.Length/2] = 0;
                } else {
                    Status = STATUS_NOT_A_DIRECTORY;
                    __leave;
                }
            }

            if (!ParentMcb) {
                if (PathName.Buffer[0] != L'\\') {
                    Status = STATUS_OBJECT_PATH_NOT_FOUND;
                    __leave;
                } else {
                    ParentMcb = Vcb->McbTree;
                }
            }

Dissecting:

            FsRtlDissectName(PathName, &RealName, &RemainName);

            if (((RemainName.Length != 0) && (RemainName.Buffer[0] == L'\\')) ||
                (RealName.Length >= 256*sizeof(WCHAR))) {
                Status = STATUS_OBJECT_NAME_INVALID;
                __leave;
            }

            if (RemainName.Length != 0) {

                PEXT2_MCB   RetMcb;

                Status = Ext2LookupFileName (
                                Vcb,
                                &RealName,
                                ParentMcb,
                                &RetMcb,
                                Inode      );

                if (!NT_SUCCESS(Status)) {
                    Status = STATUS_OBJECT_PATH_NOT_FOUND;
                    __leave;
                }

                ParentMcb = RetMcb;
                PathName  = RemainName;

                goto Dissecting;
            }

            if (FsRtlDoesNameContainWildCards(&RealName)) {
                Status = STATUS_OBJECT_NAME_INVALID;
                __leave;
            }

            ParentFcb = ParentMcb->Ext2Fcb;

            if (!ParentFcb) {

                PEXT2_INODE pTmpInode = ExAllocatePool( PagedPool, 
                                                        sizeof(EXT2_INODE));
                if (!pTmpInode) {
                    Status = STATUS_INSUFFICIENT_RESOURCES;
                    __leave;
                }

                if(!Ext2LoadInode(Vcb, ParentMcb->Inode, pTmpInode)) {
                    Status = STATUS_OBJECT_PATH_NOT_FOUND;
                    __leave;
                }

                ParentFcb = Ext2AllocateFcb(Vcb,  ParentMcb, pTmpInode);

                if (!ParentFcb) {
                    ExFreePool(pTmpInode);
                    Status = STATUS_INSUFFICIENT_RESOURCES;
                    __leave;
                }

                bParentFcbCreated = TRUE;
                ParentFcb->ReferenceCount++;
            }

            // We need to create a new one ?
            if ((CreateDisposition == FILE_CREATE ) ||
                (CreateDisposition == FILE_OPEN_IF) ||
                (CreateDisposition == FILE_OVERWRITE_IF)) {

                if (IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {
                    Status = STATUS_MEDIA_WRITE_PROTECTED;
                    __leave;
                }

                if (IsFlagOn(Vcb->Flags, VCB_WRITE_PROTECTED)) {
                    IoSetHardErrorOrVerifyDevice( IrpContext->Irp,
                                                  Vcb->Vpb->RealDevice );
                     SetFlag(Vcb->Vpb->RealDevice->Flags, DO_VERIFY_VOLUME);

                    Ext2RaiseStatus(IrpContext, STATUS_MEDIA_WRITE_PROTECTED);
                }

                if (DirectoryFile) {
                    if (TemporaryFile) {
                        Status = STATUS_INVALID_PARAMETER;
                        __leave;
                    }
                }

                if (!ParentFcb) {
                    Status = STATUS_OBJECT_PATH_NOT_FOUND;
                    __leave;
                }

                if (DirectoryFile) {
                    if (ParentFcb->Ext2Mcb->Inode == EXT2_ROOT_INO) {
                        if ( (RealName.Length == 0x10) && 
                             memcmp(RealName.Buffer, L"Recycled\0", 0x10) == 0) {
                            SetFlag( IrpSp->Parameters.Create.FileAttributes,
                                     FILE_ATTRIBUTE_READONLY );
                        }
                    }

                    Status = Ext2CreateInode( IrpContext,
                                              Vcb, 
                                              ParentFcb,
                                              EXT2_FT_DIR,
                                              IrpSp->Parameters.Create.FileAttributes,
                                              &RealName);
                } else {
                    Status = Ext2CreateInode( IrpContext,
                                              Vcb,
                                              ParentFcb,
                                              EXT2_FT_REG_FILE,
                                              IrpSp->Parameters.Create.FileAttributes,
                                              &RealName);
                }
                
                if (NT_SUCCESS(Status)) {

                    bCreated = TRUE;

                    Irp->IoStatus.Information = FILE_CREATED;                    
                    Status = Ext2LookupFileName (
                                Vcb,
                                &RealName,
                                ParentMcb,
                                &Ext2Mcb,
                                Inode      );

                    if (NT_SUCCESS(Status)) {

                        if (DirectoryFile) {
                            Ext2NotifyReportChange(
                                   IrpContext,
                                   Vcb,
                                   ParentFcb,
                                   FILE_NOTIFY_CHANGE_DIR_NAME,
                                   FILE_ACTION_ADDED );
                        } else {
                            Ext2NotifyReportChange(
                                   IrpContext,
                                   Vcb,
                                   ParentFcb,
                                   FILE_NOTIFY_CHANGE_FILE_NAME,
                                   FILE_ACTION_ADDED );
                        }
                    } else {
                        DbgBreak();
                    }
                } else {
                    DbgBreak();
                }
            } else if (OpenTargetDirectory) {
                if (IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {
                    Status = STATUS_MEDIA_WRITE_PROTECTED;
                    __leave;
                }

                if (!ParentFcb) {
                    Status = STATUS_OBJECT_PATH_NOT_FOUND;
                    __leave;
                }

                RtlZeroMemory( IrpSp->FileObject->FileName.Buffer,
                               IrpSp->FileObject->FileName.MaximumLength);
                IrpSp->FileObject->FileName.Length = RealName.Length;

                RtlCopyMemory( IrpSp->FileObject->FileName.Buffer,
                               RealName.Buffer,
                               RealName.Length );

                Fcb = ParentFcb;

                Irp->IoStatus.Information = FILE_DOES_NOT_EXIST;
                Status = STATUS_SUCCESS;
            } else {
                Status = STATUS_OBJECT_NAME_NOT_FOUND;
                __leave;
            }

        } else { // File / Dir already exists.

            if (OpenTargetDirectory) {

                if (IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {
                    Status = STATUS_MEDIA_WRITE_PROTECTED;
                    __leave;
                }

                Irp->IoStatus.Information = FILE_EXISTS;
                Status = STATUS_SUCCESS;

                RtlZeroMemory( IrpSp->FileObject->FileName.Buffer,
                               IrpSp->FileObject->FileName.MaximumLength);
                IrpSp->FileObject->FileName.Length = Ext2Mcb->ShortName.Length;

                RtlCopyMemory( IrpSp->FileObject->FileName.Buffer,
                               Ext2Mcb->ShortName.Buffer,
                               Ext2Mcb->ShortName.Length );

                //Let Mcb pointer to it's parent
                Ext2Mcb = Ext2Mcb->Parent;

                goto Openit;
            }

            // We can not create if one exists
            if (CreateDisposition == FILE_CREATE) {
                Irp->IoStatus.Information = FILE_EXISTS;
                Status = STATUS_OBJECT_NAME_COLLISION;
                __leave;
            }

            if(IsFlagOn(Ext2Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY)) {

                if ((CreateDisposition != FILE_OPEN) &&
                    (CreateDisposition != FILE_OPEN_IF)) {

                    Status = STATUS_OBJECT_NAME_COLLISION;
                    __leave;
                }

                if (NonDirectoryFile) {
                    Status = STATUS_FILE_IS_A_DIRECTORY;
                    __leave;
                }

                if (Ext2Mcb->Inode == EXT2_ROOT_INO) {

                    if (DeleteOnClose) {
                        Status = STATUS_CANNOT_DELETE;
                        __leave;
                    }

                    if (OpenTargetDirectory) {
                        Status = STATUS_INVALID_PARAMETER;
                        __leave;
                    }
                }
            }

            Irp->IoStatus.Information = FILE_OPENED;
        }

Openit:
        
        if (Ext2Mcb) {

            Fcb = Ext2Mcb->Ext2Fcb;

            if (!Fcb) {
                Fcb = Ext2AllocateFcb (Vcb, Ext2Mcb, Inode);
                bFcbAllocated = TRUE;
            }
        }
        
        if (Fcb) {

            if (IsFlagOn(Fcb->Flags, FCB_FILE_DELETED)) {
                Status = STATUS_FILE_DELETED;
                __leave;
            }

            if (FlagOn(Fcb->Flags, FCB_DELETE_PENDING)) {
                Status = STATUS_DELETE_PENDING;
                __leave;
            }

            if (bCreated) {

                //
                //  This file is just created.
                //

                if (DirectoryFile) {

                    UNICODE_STRING EntryName;
                    USHORT  NameBuf[6];

                    RtlZeroMemory(&NameBuf, 6 * sizeof(USHORT));

                    EntryName.Length = EntryName.MaximumLength = 2;
                    EntryName.Buffer = &NameBuf[0];
                    NameBuf[0] = (USHORT)'.';

                    Ext2AddEntry( IrpContext, Vcb, Fcb,
                                  EXT2_FT_DIR,
                                  Fcb->Ext2Mcb->Inode,
                                  &EntryName );

                    Ext2SaveInode( IrpContext, Vcb,
                                   Fcb->Ext2Mcb->Inode,
                                   Fcb->Inode );

                    EntryName.Length = EntryName.MaximumLength = 4;
                    EntryName.Buffer = &NameBuf[0];
                    NameBuf[0] = NameBuf[1] = (USHORT)'.';

                    Ext2AddEntry( IrpContext, Vcb, Fcb,
                                  EXT2_FT_DIR,
                                  Fcb->Ext2Mcb->Parent->Inode,
                                  &EntryName );

                    Ext2SaveInode( IrpContext, Vcb,
                                   Fcb->Ext2Mcb->Parent->Inode,
                                   ParentFcb->Inode );
                } else {
                    Status = Ext2ExpandFile(
                                IrpContext, Vcb, Fcb,
                                &(Irp->Overlay.AllocationSize));

                    if (!NT_SUCCESS(Status)) {

                        DbgBreak();

                        __leave;
                    }
                }

            } else {

                //
                //  This file alreayd exists.
                //

                if (DeleteOnClose) {

                    if (IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {
                        Status = STATUS_MEDIA_WRITE_PROTECTED;
                        __leave;
                    }

                    if (IsFlagOn(Vcb->Flags, VCB_WRITE_PROTECTED)) {
                        Status = STATUS_MEDIA_WRITE_PROTECTED;

                        IoSetHardErrorOrVerifyDevice( IrpContext->Irp,
                                                      Vcb->Vpb->RealDevice );

                        SetFlag(Vcb->Vpb->RealDevice->Flags, DO_VERIFY_VOLUME);

                        Ext2RaiseStatus(IrpContext, STATUS_MEDIA_WRITE_PROTECTED);
                    }

                    SetFlag(Fcb->Flags, FCB_DELETE_ON_CLOSE);

                } else {

                    //
                    // Just to Open file (Open/OverWrite ...)
                    //

                    if ((!IsDirectory(Fcb)) && (IsFlagOn(IrpSp->FileObject->Flags,
                                                FO_NO_INTERMEDIATE_BUFFERING))) {
                        Fcb->Header.IsFastIoPossible = FastIoIsPossible;

                        if (IsFlagOn(IrpSp->FileObject->Flags, FO_CACHE_SUPPORTED) &&
                            (Fcb->SectionObject.DataSectionObject != NULL)) {

                            if (Fcb->NonCachedOpenCount == Fcb->OpenHandleCount) {
                                /* IsFlagOn(FileObject->Flags, FO_FILE_MODIFIED) */

                                if(!IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {
                                    CcFlushCache(&Fcb->SectionObject, NULL, 0, NULL);
                                    ClearFlag(Fcb->Flags, FCB_FILE_MODIFIED);
                                }

                                CcPurgeCacheSection(&Fcb->SectionObject,
                                                     NULL,
                                                     0,
                                                     FALSE );
                            }
                        }
                    }
                }
            }

            if (!IsDirectory(Fcb)) {
                if (!IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {
                    if ((CreateDisposition == FILE_SUPERSEDE) && !IsPagingFile){
                        DesiredAccess |= DELETE;
                    } else if (((CreateDisposition == FILE_OVERWRITE) ||
                            (CreateDisposition == FILE_OVERWRITE_IF)) && !IsPagingFile) {
                        DesiredAccess |= (FILE_WRITE_DATA | FILE_WRITE_EA |
                                          FILE_WRITE_ATTRIBUTES );
                    }
                }
            }

            if (Fcb->OpenHandleCount > 0) {
                Status = IoCheckShareAccess( DesiredAccess,
                                             ShareAccess,
                                             IrpSp->FileObject,
                                             &(Fcb->ShareAccess),
                                             TRUE );

                if (!NT_SUCCESS(Status)) {
                    __leave;
                }
            } else {
                IoSetShareAccess( DesiredAccess,
                                  ShareAccess,
                                  IrpSp->FileObject,
                                  &(Fcb->ShareAccess) );
            }

            Ccb = Ext2AllocateCcb();

            Fcb->OpenHandleCount++;
            Fcb->ReferenceCount++;

            if (IsFlagOn(IrpSp->FileObject->Flags, FO_NO_INTERMEDIATE_BUFFERING)) {
                Fcb->NonCachedOpenCount++;
            }

            Vcb->OpenFileHandleCount++;
            Vcb->ReferenceCount++;
            
            IrpSp->FileObject->FsContext = (void*)Fcb;
            IrpSp->FileObject->FsContext2 = (void*) Ccb;
            IrpSp->FileObject->PrivateCacheMap = NULL;
            IrpSp->FileObject->SectionObjectPointer = &(Fcb->SectionObject);
            IrpSp->FileObject->Vpb = Vcb->Vpb;

            Status = STATUS_SUCCESS;

            Ext2Print((DBG_INFO, "Ext2CreateFile: %s OpenCount: %u ReferCount: %u\n",
                Fcb->AnsiFileName.Buffer, Fcb->OpenHandleCount, Fcb->ReferenceCount));

            if (!IsDirectory(Fcb) && !NoIntermediateBuffering ) {
                IrpSp->FileObject->Flags |= FO_CACHE_SUPPORTED;
            }

            if (!bCreated && !IsDirectory(Fcb)) {
                if ( DeleteOnClose || 
                    IsFlagOn(DesiredAccess, FILE_WRITE_DATA) || 
                    (CreateDisposition == FILE_OVERWRITE) ||
                    (CreateDisposition == FILE_OVERWRITE_IF)) {
                    if (!MmFlushImageSection( &Fcb->SectionObject,
                                              MmFlushForWrite )) {

                        Status = DeleteOnClose ? STATUS_CANNOT_DELETE :
                                                 STATUS_SHARING_VIOLATION;
                        __leave;
                    }
                }

                if ((CreateDisposition == FILE_SUPERSEDE) ||
                    (CreateDisposition == FILE_OVERWRITE) ||
                    (CreateDisposition == FILE_OVERWRITE_IF)) {

                    if (IsFlagOn(Vcb->Flags, VCB_READ_ONLY)) {
                        Status = STATUS_MEDIA_WRITE_PROTECTED;
                        __leave;
                    }

                    if (IsFlagOn(Vcb->Flags, VCB_WRITE_PROTECTED)) {

                        IoSetHardErrorOrVerifyDevice( IrpContext->Irp,
                                                      Vcb->Vpb->RealDevice );
                        SetFlag(Vcb->Vpb->RealDevice->Flags, DO_VERIFY_VOLUME);

                        Ext2RaiseStatus(IrpContext, STATUS_MEDIA_WRITE_PROTECTED);
                    }

                    Status = Ext2SupersedeOrOverWriteFile( IrpContext,
                                                  Vcb,
                                                  Fcb,
                                                  CreateDisposition );

                    if (NT_SUCCESS(Status)) {
                        __leave;
                    }

                    Status = Ext2ExpandFile(
                                IrpContext,
                                Vcb,
                                Fcb,
                                &(Irp->Overlay.AllocationSize));

                    if (!(NT_SUCCESS(Status))) {
                        __leave;
                    }

                    Ext2NotifyReportChange(
                               IrpContext,
                               Vcb,
                               Fcb,
                               FILE_NOTIFY_CHANGE_LAST_WRITE |
                               FILE_NOTIFY_CHANGE_ATTRIBUTES |
                               FILE_NOTIFY_CHANGE_SIZE,
                               FILE_ACTION_MODIFIED );


                    if (CreateDisposition == FILE_SUPERSEDE) {
                        Irp->IoStatus.Information = FILE_SUPERSEDED;
                    } else {
                        Irp->IoStatus.Information = FILE_OVERWRITTEN;
                    }
                }
            }
        }
    } __finally {

        if (FileName.Buffer)
            ExFreePool(FileName.Buffer);

        if (bParentFcbCreated) {
            ParentFcb->ReferenceCount--;
        }

        if (VcbResourceAcquired) {
            ExReleaseResourceForThreadLite(
                &Vcb->MainResource,
                ExGetCurrentResourceThread() );
        }

        if (!bFcbAllocated) {
            if (Inode)
                ExFreePool(Inode);
        } else {
            if (!Fcb && Inode)
                ExFreePool(Inode);
        }
    }
    
    return Status;
}

NTSTATUS
Ext2CreateVolume(PEXT2_IRP_CONTEXT IrpContext, PEXT2_VCB Vcb)
{
    PIO_STACK_LOCATION  IrpSp;
    PIRP                Irp;

    NTSTATUS            Status;

    ACCESS_MASK         DesiredAccess;
    ULONG               ShareAccess;

    ULONG               Options;
    BOOLEAN             DirectoryFile;
    BOOLEAN             OpenTargetDirectory;

    ULONG               CreateDisposition;

    Irp = IrpContext->Irp;
    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    Options  = IrpSp->Parameters.Create.Options;
    
    DirectoryFile = IsFlagOn(Options, FILE_DIRECTORY_FILE);
    OpenTargetDirectory = IsFlagOn(IrpSp->Flags, SL_OPEN_TARGET_DIRECTORY);

    CreateDisposition = (Options >> 24) & 0x000000ff;

    DesiredAccess = IrpSp->Parameters.Create.SecurityContext->DesiredAccess;
    ShareAccess   = IrpSp->Parameters.Create.ShareAccess;

    if (DirectoryFile) {
        return STATUS_NOT_A_DIRECTORY;
    }

    if (OpenTargetDirectory) {
        return STATUS_INVALID_PARAMETER;
    }

    if ( (CreateDisposition != FILE_OPEN) && 
         (CreateDisposition != FILE_OPEN_IF) ) {
        return STATUS_ACCESS_DENIED;
    }

    Status = STATUS_SUCCESS;

    if (Vcb->OpenHandleCount > 0) {
        Status = IoCheckShareAccess( DesiredAccess, ShareAccess,
                                     IrpSp->FileObject,
                                     &(Vcb->ShareAccess), TRUE);

        if (!NT_SUCCESS(Status)) {
            goto errorout;
        }
    } else {
        IoSetShareAccess( DesiredAccess, ShareAccess,
                          IrpSp->FileObject,
                          &(Vcb->ShareAccess)   );
    }

    if (FlagOn(DesiredAccess, FILE_READ_DATA | FILE_WRITE_DATA | FILE_APPEND_DATA)) {
        ExAcquireResourceExclusive(&Vcb->MainResource, TRUE);
        Ext2FlushFiles(Vcb, FALSE);
        Ext2FlushVolume(Vcb, FALSE);
        ExReleaseResource(&Vcb->MainResource);
    }

    {
        PEXT2_CCB   Ccb = Ext2AllocateCcb();

        if (Ccb == NULL) {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            goto errorout;
        }

        IrpSp->FileObject->Flags |= FO_NO_INTERMEDIATE_BUFFERING;
        IrpSp->FileObject->FsContext  = Vcb;
        IrpSp->FileObject->FsContext2 = Ccb;

        Vcb->ReferenceCount++;
        Vcb->OpenHandleCount++;

        Irp->IoStatus.Information = FILE_OPENED;
    }

errorout:

    return Status;
}


NTSTATUS
Ext2Create (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT      DeviceObject;
    PIRP                Irp;
    PIO_STACK_LOCATION  IrpSp;
    PEXT2_VCB           Vcb = 0;
    NTSTATUS            Status = STATUS_OBJECT_NAME_NOT_FOUND;
    PEXT2_FCBVCB        Xcb = NULL;

    DeviceObject = IrpContext->DeviceObject;

    Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;

    ASSERT(IsMounted(Vcb));
    
    Irp = IrpContext->Irp;
    
    IrpSp = IoGetCurrentIrpStackLocation(Irp);

    Xcb = (PEXT2_FCBVCB) (IrpSp->FileObject->FsContext);
    
    if (DeviceObject == Ext2Global->DeviceObject) {
        Ext2Print((DBG_INFO, "Ext2Create: Create on main device object.\n"));

        Status = STATUS_SUCCESS;
        
        Irp->IoStatus.Information = FILE_OPENED;

        Ext2UnpinRepinnedBcbs(IrpContext);

        Ext2CompleteIrpContext(IrpContext, Status);        

        return Status;
    }
   
    __try {

        if (IsFlagOn(Vcb->Flags, VCB_VOLUME_LOCKED)) {
            Status = STATUS_ACCESS_DENIED;

            if (IsFlagOn(Vcb->Flags, VCB_DISMOUNT_PENDING)) {
                Status = STATUS_VOLUME_DISMOUNTED;
            }

            __leave;
        }

        if ( ((IrpSp->FileObject->FileName.Length == 0) &&
             (IrpSp->FileObject->RelatedFileObject == NULL)) || 
             (Xcb && Xcb->Identifier.Type == EXT2VCB)  ) {
            Status = Ext2CreateVolume(IrpContext, Vcb);
        } else {
            Status = Ext2CreateFile(IrpContext, Vcb);
        }

    } __finally {

        if (!IrpContext->ExceptionInProgress)  {
            Ext2UnpinRepinnedBcbs(IrpContext);
     
            Ext2CompleteIrpContext(IrpContext, Status);
        }
    }

    return Status;
}


NTSTATUS
Ext2CreateInode(
    PEXT2_IRP_CONTEXT   IrpContext,
    PEXT2_VCB           Vcb,
    PEXT2_FCB           ParentFcb,
    ULONG               Type,
    ULONG               FileAttr,
    PUNICODE_STRING     FileName)
{
    NTSTATUS    Status;
    ULONG       Inode;
    ULONG       Group;

    EXT2_INODE  Ext2Ino;

    RtlZeroMemory(&Ext2Ino, sizeof(EXT2_INODE));

    Group = (ParentFcb->Ext2Mcb->Inode - 1) / BLOCKS_PER_GROUP;

    Ext2Print(( DBG_INFO,
                "Ext2CreateInode: %S in %S(Inode=%xh)\n",
                FileName->Buffer, 
                ParentFcb->Ext2Mcb->ShortName.Buffer, 
                ParentFcb->Ext2Mcb->Inode ));

    Status = Ext2NewInode(IrpContext, Vcb, Group,Type, &Inode);

    if (!NT_SUCCESS(Status)) {
        goto errorout;
    }

    Status = Ext2AddEntry(IrpContext, Vcb, ParentFcb, Type, Inode, FileName);

    if (!NT_SUCCESS(Status)) {
        DbgBreak();
        Ext2FreeInode(IrpContext, Vcb, Inode, Type);

        goto errorout;
    }

    Ext2SaveInode(IrpContext, Vcb, ParentFcb->Ext2Mcb->Inode, ParentFcb->Inode);

    Ext2Ino.i_ctime = ParentFcb->Inode->i_mtime;
    Ext2Ino.i_mode =  ( S_IPERMISSION_MASK &
                        ParentFcb->Inode->i_mode );
    Ext2Ino.i_uid = ParentFcb->Inode->i_uid;
    Ext2Ino.i_gid = ParentFcb->Inode->i_gid;

    Ext2Ino.i_dir_acl = ParentFcb->Inode->i_dir_acl;
    Ext2Ino.i_file_acl = ParentFcb->Inode->i_file_acl;
    Ext2Ino.i_generation = ParentFcb->Inode->i_generation;

    Ext2Ino.osd2 = ParentFcb->Inode->osd2;

    if (IsFlagOn(FileAttr, FILE_ATTRIBUTE_READONLY)) {
        Ext2SetReadOnly(Ext2Ino.i_mode);
    }

    if (Type == EXT2_FT_DIR)  {
        Ext2Ino.i_mode |= S_IFDIR;
        Ext2Ino.i_links_count = 2;
    } else if (Type == EXT2_FT_REG_FILE) {
        Ext2Ino.i_mode |= S_IFREG;
        Ext2Ino.i_links_count = 1;
    } else {
        DbgBreak();
        Ext2Ino.i_links_count = 1;
    }

    Ext2SaveInode(IrpContext, Vcb, Inode, &Ext2Ino);

    Ext2Print((DBG_INFO, "Ext2CreateInode: New Inode = %xh (Type=%xh)\n", Inode, Type));
            
errorout:

    return Status;
}


NTSTATUS
Ext2SupersedeOrOverWriteFile(
        PEXT2_IRP_CONTEXT IrpContext,
        PEXT2_VCB Vcb,
        PEXT2_FCB Fcb,
        ULONG     Disposition)
{
    LARGE_INTEGER   CurrentTime;
    LARGE_INTEGER   AllocationSize;
    NTSTATUS        Status = STATUS_SUCCESS;
    
    KeQuerySystemTime(&CurrentTime);

    AllocationSize.QuadPart = (LONGLONG)0;

    if (!MmCanFileBeTruncated(&(Fcb->SectionObject), &(AllocationSize))) {
        Status = STATUS_USER_MAPPED_FILE;
        return Status;
    }

    Status = Ext2TruncateFile(IrpContext, Vcb, Fcb, &AllocationSize);

    if (NT_SUCCESS(Status)) {

        Fcb->Header.AllocationSize.QuadPart = 
        Fcb->Header.FileSize.QuadPart =  (LONGLONG) 0;

        Fcb->Inode->i_size = 0;

        if (S_ISREG(Fcb->Inode->i_mode)) {
            Fcb->Inode->i_size_high = 0;
        }

        if (Disposition == FILE_SUPERSEDE)
            Fcb->Inode->i_ctime = Ext2InodeTime(CurrentTime);

        Fcb->Inode->i_atime =
        Fcb->Inode->i_mtime = Ext2InodeTime(CurrentTime);
    } else {
        LARGE_INTEGER   iSize;

        iSize.QuadPart  =  (LONGLONG) Fcb->Inode->i_size;

        if (S_ISREG(Fcb->Inode->i_mode))
            iSize.HighPart = (LONG)(Fcb->Inode->i_size_high);

        if (iSize.QuadPart > Fcb->Header.AllocationSize.QuadPart)
            iSize.QuadPart = Fcb->Header.AllocationSize.QuadPart;
    
        Fcb->Header.FileSize.QuadPart =  iSize.QuadPart;

        Fcb->Inode->i_size = iSize.LowPart;
        Fcb->Inode->i_size_high = (ULONG) iSize.HighPart;
    }

    Ext2SaveInode(IrpContext, Vcb, Fcb->Ext2Mcb->Inode, Fcb->Inode);

    return Status;
}
